from typing import Dict, Type

import numpy as np
import torch
import torch.nn.functional as F
from einops.layers.torch import Rearrange
from torch import nn

BLOCK_REGISTRY: Dict[str, Type[nn.Module]] = {}

ACTIVATION_LAYER_REGISTRY: Dict[str, Type[nn.Module]] = {
    "relu": nn.ReLU,
    "gelu": nn.GELU,
}

KERAS_LAYER_DEFAULTS = {
    "BatchNorm1d": {
        "momentum": 1 - 0.99,  # momentum is 1-momentum
        "eps": 1e-3,
    },
}


class BasenjiSoftplus(nn.Module):
    def __init__(self, exp_max=10000):
        super().__init__()
        self.exp_max = exp_max

    def forward(self, x: torch.Tensor):
        x.clip_(-self.exp_max, self.exp_max)
        return F.softplus(x)


ACTIVATION_LAYER_REGISTRY["softplus"] = BasenjiSoftplus


class Residual(nn.Module):
    def __init__(self, module):
        super().__init__()
        self.module = module

    def forward(self, x):
        return self.module(x) + x


class Cropping1d(nn.Module):
    def __init__(self, cropping: int, **kwargs):
        super().__init__()
        self.cropping = cropping

    def forward(self, x):
        return x[..., self.cropping : -self.cropping]


BLOCK_REGISTRY["Cropping1D"] = Cropping1d


class KerasMaxPool1d(nn.Module):
    def __init__(
        self,
        pool_size=2,
        padding="valid",
        dilation=1,
        return_indices=False,
        ceil_mode=False,
    ):
        """Hard coded to only be compatible with kernel size and stride of 2."""
        super().__init__()
        self.padding = padding
        _padding = 0
        if pool_size != 2:
            raise NotImplementedError("MaxPool1D with kernel size other than 2.")
        self.pool = nn.MaxPool1d(
            kernel_size=pool_size,
            padding=_padding,
            dilation=dilation,
            return_indices=return_indices,
            ceil_mode=ceil_mode,
        )

    def forward(self, x):
        # (b c h)
        if self.padding == "same" and x.shape[-1] % 2 == 1:
            x = F.pad(x, (0, 1), value=-float("inf"))
        return self.pool(x)


class BasenjiConvBlock(nn.Module):
    def __init__(
        self,
        in_channels,
        filters,
        kernel_size=1,
        activation="relu",
        stride=1,
        dilation_rate=1,
        dropout=0,
        pool_size=1,
        norm_type=None,
        bn_momentum=0.99,
        residual=False,
        padding="same",
    ):
        super().__init__()

        block = nn.ModuleList()

        # activation
        block.append(ACTIVATION_LAYER_REGISTRY[activation]())

        # conv
        block.append(
            nn.Conv1d(
                in_channels=in_channels,
                out_channels=filters,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=int(round(dilation_rate)),
                bias=(norm_type is None),
            )
        )
        # batch norm
        if norm_type == "batch":
            # Use Keras default eps
            block.append(nn.BatchNorm1d(filters, momentum=1 - bn_momentum, eps=1e-3))
        else:
            raise NotImplementedError(f"BasenjiConvBlock with norm_type = {norm_type}")
        # dropout
        if dropout > 0:
            block.append(nn.Dropout(p=dropout))
        # residual
        if residual:
            block = nn.ModuleList([Residual(nn.Sequential(*block))])
        # pool
        if pool_size > 1:
            block.append(KerasMaxPool1d(pool_size, padding))

        self.block = nn.Sequential(*block)
        self.out_channels = filters

    def forward(self, x):
        return self.block(x)


BLOCK_REGISTRY["conv_block"] = BasenjiConvBlock


class BasenjiConvTower(nn.Module):
    def __init__(
        self,
        in_channels,
        filters_init,
        filters_end=None,
        filters_mult=None,
        divisible_by=1,
        repeat=1,
        **kwargs,
    ):
        super().__init__()

        def _round(x):
            return int(np.round(x / divisible_by) * divisible_by)

        # determine multiplier
        if filters_mult is None:
            assert filters_end is not None
            filters_mult = np.exp(np.log(filters_end / filters_init) / (repeat - 1))

        rep_filters = filters_init
        in_channels = in_channels
        tower = nn.ModuleList()
        for _ in range(repeat):
            tower.append(
                BasenjiConvBlock(
                    in_channels=in_channels, filters=_round(rep_filters), **kwargs
                )
            )
            in_channels = _round(rep_filters)
            rep_filters *= filters_mult

        self.tower = nn.Sequential(*tower)
        self.out_channels = in_channels

    def forward(self, x):
        return self.tower(x)


BLOCK_REGISTRY["conv_tower"] = BasenjiConvTower


class BasenjiDilatedResidual(nn.Module):
    def __init__(
        self,
        in_channels,
        filters,
        kernel_size=3,
        rate_mult=2,
        dropout=0,
        repeat=1,
        norm_type=None,
        round=False,
        **kwargs,
    ):
        super().__init__()
        dilation_rate = 1
        in_channels = in_channels
        block = nn.ModuleList()
        for _ in range(repeat):
            inner_block = nn.ModuleList()

            inner_block.append(
                BasenjiConvBlock(
                    in_channels=in_channels,
                    filters=filters,
                    kernel_size=kernel_size,
                    dilation_rate=int(np.round(dilation_rate)),
                    norm_type=norm_type,
                    **kwargs,
                )
            )

            inner_block.append(
                BasenjiConvBlock(
                    in_channels=filters,
                    filters=in_channels,
                    dropout=dropout,
                    norm_type=norm_type,
                    **kwargs,
                )
            )

            if norm_type is None:
                raise NotImplementedError(
                    "basenji.layers.scale is not ported to PyTorch."
                )

            block.append(Residual(nn.Sequential(*inner_block)))

            dilation_rate *= rate_mult
            if round:
                dilation_rate = np.round(dilation_rate)
        self.block = nn.Sequential(*block)
        self.out_channels = in_channels

    def forward(self, x):
        return self.block(x)

BLOCK_REGISTRY["dilated_residual"] = BasenjiDilatedResidual


class BasenjiFinal(nn.Module):
    def __init__(
        self, in_features, units, activation='linear', flatten=False, **kwargs
    ) -> None:
        super().__init__()
        block = nn.ModuleList()
        if flatten:
            block.append(Rearrange('b ... -> b (...)'))
        else:
            # in this case, same as pointwise (1x1) convolution
            block.append(Rearrange('b c l -> b l c'))
        block.append(nn.Linear(in_features=in_features, out_features=units))
        block.append(ACTIVATION_LAYER_REGISTRY[activation]())
        self.block = nn.Sequential(*block)
    
    def forward(self, x):
        return self.block(x)

BLOCK_REGISTRY["final"] = BasenjiFinal


import torch
from tensorflow import keras
import einops as ein

def copy_dense(tf_dense, torch_dense, has_bias=True):
    tf_weight, tf_bias = tf_dense.get_weights()
    torch_dense.weight.data.copy_(torch.from_numpy(ein.rearrange(tf_weight, 'i o -> o i')))
    if has_bias: torch_dense.bias.data.copy_(torch.from_numpy(tf_bias))

def copy_conv(tf_conv, torch_conv):
    tf_weight = ein.rearrange(tf_conv.get_weights()[0], 'k i o -> o i k')
    torch_conv.weight.data.copy_(torch.from_numpy(tf_weight))

def copy_bn(tf_bn, torch_bn):
    tf_gamma, tf_beta, tf_ema_mean, tf_ema_var = map(torch.from_numpy, tf_bn.get_weights())
    torch_bn.weight.data.copy_(tf_gamma)
    torch_bn.bias.data.copy_(tf_beta)
    torch_bn.running_mean.data.copy_(tf_ema_mean)
    torch_bn.running_var.data.copy_(tf_ema_var)

def copy_tf_to_pytorch(tf_basenji, torch_basenji):
    # both models are nested sequential with the same layers
    tf_conv_layers = [module for module in tf_basenji.model.layers if isinstance(module, keras.layers.Conv1D)]
    torch_conv_layers = [module for module in torch_basenji.modules() if isinstance(module, torch.nn.Conv1d)]
    for tf_conv, torch_conv in zip(tf_conv_layers, torch_conv_layers):
        copy_conv(tf_conv, torch_conv)
    
    tf_bn_layers = [module for module in tf_basenji.model.layers if isinstance(module, keras.layers.BatchNormalization)]
    torch_bn_layers = [module for module in torch_basenji.modules() if isinstance(module, torch.nn.BatchNorm1d)]
    for tf_bn, torch_bn in zip(tf_bn_layers, torch_bn_layers):
        copy_bn(tf_bn, torch_bn)
        
    tf_dense_layers = [module for module in tf_basenji.model.layers if isinstance(module, keras.layers.Dense)]
    torch_dense_layers = [module for module in torch_basenji.modules() if isinstance(module, torch.nn.Linear)]
    for tf_dense, torch_dense in zip(tf_dense_layers, torch_dense_layers):
        copy_dense(tf_dense, torch_dense)

    print('success')


"""Re-implementation of Basenji2, [Cross-species regulatory sequence activity prediction](https://doi.org/10.1371/journal.pcbi.1008050).

This is a minimal re-implementation that is not compatible with the full range of configuration
supported by the Basenji GitHub to instantiate a `basenji.seqnn.SeqNN` object.
"""

from __future__ import annotations

import warnings
from collections import OrderedDict
from textwrap import dedent
from typing import Any, Dict, List, Optional

import numpy as np
import pytorch_lightning as pl
import torch
import torchinfo
from natsort import natsorted
from pooch import retrieve
from pytorch_lightning.callbacks import EarlyStopping, ModelCheckpoint
from torch import nn
from torchmetrics import Metric, MetricCollection, PearsonCorrCoef

from .blocks import ACTIVATION_LAYER_REGISTRY, BLOCK_REGISTRY


def basenji2_weights() -> str:
    return retrieve(
        url="https://zenodo.org/records/14969593/files/basenji2.pth?download=1",
        known_hash="md5:1f46acc933e05515faf4d7eec0e27d42",
    )


basenji2_params = {
    "train": {
        "batch_size": 4,
        "optimizer": "sgd",
        "learning_rate": 0.15,
        "momentum": 0.99,
        "patience": 16,
        "clip_norm": 2,
    },
    "model": {
        "seq_length": 131072,
        "target_length": 1024,
        "activation": "gelu",
        "norm_type": "batch",
        "bn_momentum": 0.9,
        "trunk": [
            {"name": "conv_block", "filters": 288, "kernel_size": 15, "pool_size": 2},
            {
                "name": "conv_tower",
                "filters_init": 339,
                "filters_mult": 1.1776,
                "kernel_size": 5,
                "pool_size": 2,
                "repeat": 6,
            },
            {
                "name": "dilated_residual",
                "filters": 384,
                "rate_mult": 1.5,
                "repeat": 11,
                "dropout": 0.3,
                "round": True,
            },
            {"name": "Cropping1D", "cropping": 64},
            {"name": "conv_block", "filters": 1536, "dropout": 0.05},
        ],
        "head_human": {"name": "final", "units": 5313, "activation": "softplus"},
    },
}


# 30,123,584 parameters
# 30,099,228 trainable parameters
class Basenji2(nn.Module):
    def __init__(self, params: Dict) -> None:
        super().__init__()
        for k, v in params.items():
            self.__setattr__(k, v)
        self.build_model()

    def build_block(self, block_params: Dict[str, Any]):
        block_args = {}
        block_name = block_params["name"]

        # set global defaults
        global_vars = ["activation", "bn_momentum", "norm_type", "padding"]
        for gv in global_vars:
            gv_value = getattr(self, gv, False)
            if gv_value:
                block_args[gv] = gv_value

        # set remaining params
        block_args.update(block_params)
        del block_args["name"]

        block = BLOCK_REGISTRY[block_name]
        return block(**block_args)

    def build_model(self):
        model_trunk = nn.ModuleList()

        CONV_LAYERS = frozenset(["conv_block", "conv_tower", "dilated_residual"])

        # trunk
        in_channels = 4
        for block_params in self.trunk:
            if block_params["name"] in CONV_LAYERS:
                block_params["in_channels"] = in_channels

            block = self.build_block(block_params)
            model_trunk.append(block)

            if block_params["name"] in CONV_LAYERS:
                in_channels = block.out_channels

        # final trunk activation
        model_trunk.append(ACTIVATION_LAYER_REGISTRY[self.activation]())
        trunk = nn.Sequential(*model_trunk)

        # head
        trunk_stats = torchinfo.summary(
            trunk, input_size=(1, 4, self.seq_length), verbose=0
        )
        trunk_out_features = trunk_stats.summary_list[-1].output_size[-2]

        head_keys = natsorted([v for v in vars(self) if v.startswith("head")])
        self.heads = [getattr(self, hk) for hk in head_keys]
        _heads = nn.ModuleDict()
        for head in self.heads:
            if not isinstance(head, list):
                head = [head]

            # build blocks
            for block_params in head:
                block_params["in_features"] = trunk_out_features
                _heads[block_params["name"]] = self.build_block(block_params)

        head = nn.Sequential(OrderedDict(_heads))

        # finish
        self.model = nn.Sequential(OrderedDict([("trunk", trunk), ("head", head)]))

    def forward(self, x, return_only_embeddings=False):
        if return_only_embeddings:
            return self.model.trunk(x)
        else:
            return self.model(x)


class PLBasenji2(pl.LightningModule):
    def __init__(
        self,
        params: Optional[Dict] = None,
        pretrained=False,
        head=True,
        metrics: Optional[List[Metric]] = None,
    ) -> None:
        super().__init__()
        self.save_hyperparameters()

        if params is None:
            params = basenji2_params
        model_params = params["model"]

        if not head and params is not None and not pretrained:
            warnings.warn(
                "Got head=False and non-null train parameters. Did you mean to train a headless model?"
            )

        if not head:
            model_params.pop("head_human", None)

        if params is None and not pretrained:
            self.hparams["train_params"] = params["train"]

        if not pretrained:
            self.hparams["lr"] = self.hparams.train_params["learning_rate"]  # type: ignore
            self.hparams["batch_size"] = self.hparams.train_params["batch_size"]  # type: ignore
            msg = dedent(
                """
                Use PLBasenji2.get_cross2020_trainer(...) to get a trainer matching the Basenji2 paper.
                Also use a batch size of 4 with your dataloader/datamodule, or adjust the learning
                rate proportionally. Original parameters: lr=0.15, batch_size=4
                """
            ).strip()
            print(msg)

        self.basenji2 = Basenji2(model_params)
        self.loss_fn = nn.PoissonNLLLoss(log_input=False)

        if pretrained:
            self.basenji2.load_state_dict(torch.load(basenji2_weights()), strict=False)
        else:

            @torch.no_grad()
            def init_weights(m):
                if isinstance(m, (nn.Conv1d, nn.Linear)):
                    nn.init.kaiming_normal_(
                        m.weight, nonlinearity="relu"
                    )  # matches Keras, gain of sqrt(2) regardless of activation function
                    if getattr(m, "bias", False):
                        m.bias.fill_(0)

            self.basenji2.apply(init_weights)

        # TODO: allow different input sizes (not needed atm)
        model_stat = torchinfo.summary(self.basenji2, (1, 4, 131072), verbose=0)
        out_shape = model_stat.summary_list[-1].output_size

        if metrics is None:
            metrics = []
        metrics.append(PearsonCorrCoef(num_outputs=np.prod(out_shape)))
        _metrics = MetricCollection(metrics)
        self.train_metrics = _metrics.clone(prefix="train/")
        self.val_metrics = _metrics.clone(prefix="val/")
        self.test_metrics = _metrics.clone(prefix="test/")

    def forward(self, x):
        return self.basenji2(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        x = self.basenji2(x)
        loss = self.loss_fn(x, y)
        self.log("train/loss", loss)
        self.log_dict(self.train_metrics(x, y))
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch
        x = self.basenji2(x)
        loss = self.loss_fn(x, y)
        self.log("val/loss", loss)
        self.log_dict(self.val_metrics(x, y))

    def test_step(self, batch, batch_idx):
        x, y = batch
        x = self.basenji2(x)
        self.log_dict(self.test_metrics(x, y))

    def predict_step(self, batch, batch_idx, dataloader_idx=0):
        x = batch
        return self.basenji2(x)

    def configure_optimizers(self) -> torch.optim.Optimizer:
        if self.hparams.train_params["optimizer"] == "sgd":  # type: ignore
            optim = torch.optim.SGD(
                self.parameters(),
                lr=self.hparams.lr,  # type: ignore
                momentum=self.hparams.train_params["momentum"],  # type: ignore
            )
        else:
            raise NotImplementedError("Optimizer isn't SGD")
        return optim

    @classmethod
    def get_cross2020_trainer(cls, checkpoint_dir, **kwargs) -> pl.Trainer:
        callbacks = [
            EarlyStopping("val/PearsonCorrCoef", mode="max", patience=16),
            ModelCheckpoint(
                dirpath=checkpoint_dir, filename="{epoch}-{val_PearsonCorrCoef:.2f}"
            ),
        ]
        return pl.Trainer(
            callbacks=callbacks,
            gradient_clip_val=2,
            gradient_clip_algorithm="norm",
            **kwargs,
        )
